import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def plot_index(df: pd.DataFrame, date_col: str, precip_col: str, save_file: str=None,
               index_type: str='SPI', bin_width: int=22):

    pos_index = df.loc[df[precip_col] >= 0]
    neg_index = df.loc[df[precip_col] < 0]

    fig, ax = plt.subplots()
    ax.bar(pos_index[date_col], pos_index[precip_col], width=bin_width, align='center', color='b')
    ax.bar(neg_index[date_col], neg_index[precip_col], width=bin_width, align='center', color='r')
    ax.grid(True)
    ax.set_xlabel("Date")
    ax.set_ylabel(index_type)

    if save_file:
        plt.savefig(save_file, dpi=400)

    return fig

def best_fit_distribution(data: np.array, dist_list: list, fit_type: str='lmom', bins: int=10,
                          save_file: str=None, **kwargs):
    '''
    Method to find the best distribution for observational data. Calculates the Sum of the
    Squares error between fitted distribution and pdf.
    Inspired by: http://stackoverflow.com/questions/6620471/fitting-empirical-distribution-to-theoretical-ones-with-scipy-python

    Parameters
    ----------
    data: np.array size: [Number Observations, ]
        A numpy array of size [Number Observations, ] with the precipiation data.

    dist_type: list
        The distribution type to fit using either L-moments or MLE
            'gam' - Gamma
            'exp' - Exponential
            'gev' - Generalised Extreme Value
            'gpa' - Generalised Pareto
            'gum' - Gumbel
            'nor' - Normal
            'pe3' - Pearson III
            'wei' - Weibull

        The distribution type to fit using ONLY MLE
            'glo' - Generalised Logistic
            'gno' - Generalised Normal
            'kap' - Kappa

        The distribution type to fit using ONLY L-moments
            'wak' - Wakeby


    fit_type: str ("lmom" or "mle")
        Specify the type of fit to use for fitting distribution to the precipitation data. Either
        L-moments (lmom) or Maximum Likelihood Estimation (mle). Note use L-moments when comparing
        to NCAR's NCL code and R's packages to calculate SPI and SPEI.

    bins: int
        Number of bins to bin precipitation data

    save_file: str
        File path and name to save figure of precipitation data and fitted distributions.

    Returns
    -------
    sse: dict (key - distribution, value - sum of square error)
        The sum of the squares error between fitted distribution and pdf.
    '''
    y, x = np.histogram(data, bins=bins, normed=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0

    sse = {}
    fig, ax = plt.subplots()
    ax.bar(x, y, width=0.5, align='center', color='b', alpha=0.5, label='data')

    for dist_name in dist_list:
        distrb = getattr(distr, dist_name)

        if fit_type == 'lmom':
            params = distrb.lmom_fit(data, **kwargs)

        elif fit_type == 'mle':
            params = distrb.fit(data, **kwargs)

        else:
            raise AttributeError(f"{fit_type} is not an option. Option fit_types are mle and lmom")

        pdf = distrb.pdf(x, **params)
        sse[dist_name] = np.sum((y - pdf)**2)
        ax.plot(x, pdf, label=dist_name)

    ax.legend()
    ax.grid(True)

    if save_file:
        plt.savefig(save_file, dpi=400)
    else:
        plt.show()

    sse = sorted(sse.items(), key=lambda x: x[1], reverse=False)
    return sse




# -*- coding: utf-8 -*-

# lmoments3 library
# Copyright (C) 2012, 2014  J. R. M. Hosking, William Asquith,
# Sam Gillespie, Pierre Gérard-Marchant, Florenz A. P. Hollebrandse
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import distr
import math


def AIC(data, distr_name, distr_paras):
    distr_f = getattr(distr, distr_name.lower())  # scipy rv_continous class

    NLL = distr_f.nnlf(data, **distr_paras)
    k = distr_f.numargs + 2  # Include location and scale in addition to shape parameters
    AIC = 2 * k + 2 * NLL
    return AIC


def AICc(data, distr_name, distr_paras):
    distr_f = getattr(distr, distr_name.lower())  # scipy rv_continous class

    AICbase = AIC(data, distr_name, distr_paras)
    k = distr_f.numargs + 2  # Include location and scale in addition to shape parameters
    diff = 2 * k * (k + 1) / (len(data) - k - 1)
    AICc = AICbase + diff
    return AICc


def BIC(data, distr_name, distr_paras):
    distr_f = getattr(distr, distr_name.lower())  # scipy rv_continous class

    NLL = distr_f.nnlf(data, **distr_paras)
    k = distr_f.numargs + 2  # Include location and scale in addition to shape parameters
    BIC = k * math.log(len(data)) + 2 * NLL
    return BIC



# -*- coding: utf-8 -*-

"""
This file contains a Python implimentation of the lmoments.f library created by
J. R. M. HOSKING.
The base Fortran code is copyright of the IBM Corperation, and the licensing
information is shown below:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IBM software disclaimer
LMOMENTS: Fortran routines for use with the method of L-moments
Permission to use, copy, modify and distribute this software for any purpose
and without fee is hereby granted, provided that this copyright and permission
notice appear on all copies of the software. The name of the IBM Corporation
may not be used in any advertising or publicity pertaining to the use of the
software. IBM makes no warranty or representations about the suitability of the
software for any purpose. It is provided "AS IS" without any express or implied
warranty, including the implied warranties of merchantability, fitness for a
particular purpose and non-infringement. IBM shall not be liable for any direct,
indirect, _special or consequential damages resulting from the loss of use,
data or projects, whether in an action of contract or tort, arising out of or
in connection with the use or performance of this software.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Additional code from the R library "lmomco" has been converted into Python.
This library was developed by WILLIAM ASQUITH, and was released under the GPL-3
License. Copyright (C) 2012 WILLIAM ASQUITH
The Python translation was conducted by:
    Sam Gillespie
    Numerical Analyst
    C&R Consulting
    Townsville Australia
    September 2013
For more information, or to report bugs, contact:
    sam@candrconsulting.com.au
Licensing for Python Translation:
####################################################
    Copyright (C) 2014 Sam Gillespie
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.Version 0.1.0:
####################################################
"""
import scipy.special
import numpy as np


def lmom_ratios_fn(data, nmom=5):
    """
    Estimate `nmom` number of L-moments from a sample `data`.
    :param data: Sequence of (sample) data
    :type data: list or array-like sequence
    :param nmom: number of L-moments to estimate
    :type nmom: int
    :return: L-moment ratios like this: l1, l2, t3, t4, t5, .. . As in: items 3 and higher are L-moment ratios.
    :rtype: list
    """

    if nmom <= 5:
        return _samlmusmall(data, nmom)
    else:
        return _samlmularge(data, nmom)


def _samlmularge(x, nmom=5):
    try:
        x = np.asarray(x, dtype=np.float64)
        n = len(x)
        x.sort()
    except ValueError:
        raise ValueError("Input data to estimate L-moments must be numeric.")

    if nmom <= 0:
        raise ValueError("Invalid number of sample L-moments")

    if n < nmom:
        raise ValueError("Insufficient length of data for specified nmoments")

    ##Calculate first order
    l = [np.sum(x) / scipy.special.comb(n, 1, exact=True)]

    if nmom == 1:
        return l[0]

    #Setup comb table, where comb[i][x] refers to comb(x,i)
    comb = []
    for i in range(1, nmom):
        comb.append([])
        for j in range(n):
            comb[-1].append(scipy.special.comb(j, i, exact=True))

    for mom in range(2, nmom + 1):
        coefl = 1.0 / mom * 1.0 / scipy.special.comb(n, mom, exact=True)
        xtrans = []
        for i in range(0, n):
            coeftemp = []
            for j in range(0, mom):
                coeftemp.append(1)

            for j in range(0, mom - 1):
                coeftemp[j] = coeftemp[j] * comb[mom - j - 2][i]

            for j in range(1, mom):
                coeftemp[j] = coeftemp[j] * comb[j - 1][n - i - 1]

            for j in range(0, mom):
                coeftemp[j] = coeftemp[j] * scipy.special.comb(mom - 1, j, exact=True)

            for j in range(0, int(0.5 * mom)):
                coeftemp[j * 2 + 1] = -coeftemp[j * 2 + 1]
            coeftemp = sum(coeftemp)
            xtrans.append(x[i] * coeftemp)

        if mom > 2:
            l.append(coefl * sum(xtrans) / l[1])
        else:
            l.append(coefl * sum(xtrans))
    return l


def _samlmusmall(x, nmom=5):
    try:
        x = np.asarray(x, dtype=np.float64)
        n = len(x)
        x.sort()
    except ValueError:
        raise ValueError("Input data to estimate L-moments must be numeric.")

    if nmom <= 0 or nmom > 5:
        raise ValueError("Invalid number of sample L-moments")

    if n < nmom:
        raise ValueError("Insufficient length of data for specified nmoments")

    # First L-moment

    l1 = np.sum(x) / scipy.special.comb(n, 1, exact=True)

    if nmom == 1:
        return l1

    # Second L-moment

    comb1 = range(n)
    coefl2 = 0.5 / scipy.special.comb(n, 2, exact=True)
    sum_xtrans = sum([(comb1[i] - comb1[n - i - 1]) * x[i] for i in range(n)])
    l2 = coefl2 * sum_xtrans

    if nmom == 2:
        return [l1, l2]

    # Third L-moment

    comb3 = [scipy.special.comb(i, 2, exact=True) for i in range(n)]
    coefl3 = 1.0 / 3.0 / scipy.special.comb(n, 3, exact=True)
    sum_xtrans = sum([(comb3[i] - 2 * comb1[i] * comb1[n - i - 1] + comb3[n - i - 1]) * x[i] for i in range(n)])
    l3 = coefl3 * sum_xtrans / l2

    if nmom == 3:
        return [l1, l2, l3]

    # Fourth L-moment

    comb5 = [scipy.special.comb(i, 3, exact=True) for i in range(n)]
    coefl4 = 0.25 / scipy.special.comb(n, 4, exact=True)
    sum_xtrans = sum(
        [(comb5[i] - 3 * comb3[i] * comb1[n - i - 1] + 3 * comb1[i] * comb3[n - i - 1] - comb5[n - i - 1]) * x[i]
         for i in range(n)])
    l4 = coefl4 * sum_xtrans / l2

    if nmom == 4:
        return [l1, l2, l3, l4]

    # Fifth L-moment

    comb7 = [scipy.special.comb(i, 4, exact=True) for i in range(n)]
    coefl5 = 0.2 / scipy.special.comb(n, 5, exact=True)
    sum_xtrans = sum(
        [(comb7[i] - 4 * comb5[i] * comb1[n - i - 1] + 6 * comb3[i] * comb3[n - i - 1] -
          4 * comb1[i] * comb5[n - i - 1] + comb7[n - i - 1]) * x[i]
         for i in range(n)])
    l5 = coefl5 * sum_xtrans / l2

    return [l1, l2, l3, l4, l5]



import setuptools
from setuptools import setup

setup(name='standard_precip',
      version='1.0',
      description='Functions to calculate SPI and SPEI',
      url='http://github.com/e-baumer/standard_precip',
      author='Eric Nussbaumer',
      author_email='ebaumer@gmail.com',
      # packages=setuptools.find_packages(),
      license='Apache License 2.0',
      packages=['standard_precip', 'standard_precip/lmoments'],
      install_requires=[
        'numpy>=1.19.0',
        'matplotlib>=3.1.2',
        'scipy>=1.5.4',
        'pandas>=1.1.5',
      ],
      python_requires='>=3.6',
      zip_safe=False)



from functools import reduce

import numpy as np
import pandas as pd
import scipy.stats as scs
import matplotlib.pyplot as plt

from standard_precip.lmoments import distr


class BaseStandardIndex():
    '''
    Calculate the SPI or SPEI index. A user specified distribution is fit to the precip data.
    The CDF of this distribution is then calculated after which the the standard normal
    distribution is calculated which gives the index. A distribution can be fit over the
    precipitation data either using MLE or L-moments. NCAR's SPI calculators and the SPI and
    SPEI R packages both use L-moments to fit the distribution. There are advantages and
    disadvantages to each technique.

    This calculation can be done on any time scale. Built in temporal scales include daily,
    weekly, and monthly; however, the user can define their own timescale.

    One should put some thought into the type of distribution fit to the
    data. Precipitation can have zero value and some distributions are only
    defined over interval (0, inf). Python's gamma distribution is defined
    over [0, inf). In addition SPEI which is constructed from precipitation
    - PET or (P-PET) can take on negative values.
    '''

    def __init__(self):
        self.distrb = None
        self.non_zero_distr = ['gam', 'pe3']
        self._df_copy = None
        self.freq_col = None

    @staticmethod
    def rolling_window_sum(df: pd.DataFrame, precip_cols: list, span: int=1, window_type: str=None,
                           center: bool=False, **kwargs):
        '''
        This is a helper method which will find the rolling sum of precipitation data.
        '''
        precip_cols_new = []
        for p in precip_cols:
            new_col_name = p+f"_scale_{span}"
            df[new_col_name] = df[p].rolling(
                window=span, win_type=window_type, center=center, **kwargs
            ).sum()
            precip_cols_new.append(new_col_name)

        return df, precip_cols_new

    @staticmethod
    def check_duplicate_dates(df, date_col):
        '''
        Method to check duplicate dates in dataframe. If duplicates are found, the row corresponding
        to the first date found is used.
        '''
        if df.duplicated(subset=date_col).any():
            print("Found duplicate dates in dataframe. Removing duplicates and using first date found")
            df = df.drop_duplicates(subset=date_col)

        return df

    def fit_distribution(self, data: np.array, dist_type: str, fit_type: str='lmom', **kwargs):
        '''
        Fit given distribution to historical precipitation data.
        The fit is accomplished using either L-moments or MLE (Maximum Likelihood Estimation).

        For distributions that use the Gamma Function (Gamma and Pearson 3) remove observations
        that have 0 precipitation values and fit using non-zero observations. Also find probability
        of zero observation (estimated by number of zero obs / total obs). This is for latter use
        in calculating the CDF using (Thom, 1966. Some Methods of Climatological Analysis)
        '''

        # Get distribution type
        self.distrb = getattr(distr, dist_type)

        # Determine zeros if distribution can not handle x = 0
        p_zero = None
        if dist_type in self.non_zero_distr:
            p_zero = data[data == 0].shape[0] / data.shape[0]
            data = data[data != 0]

        if (data.shape[0]<4) or (p_zero==1):
            params = None

        else:
            # Fit distribution
            if fit_type == 'lmom':
                params = self.distrb.lmom_fit(data, **kwargs)

            elif fit_type == 'mle':
                params = self.distrb.fit(data, **kwargs)

            else:
                raise AttributeError(f"{fit_type} is not an option. Option fit_types are mle and lmom")

        return params, p_zero

    def cdf_to_ppf(self, data, params, p_zero):
        '''
        Take the specific distributions fitted parameters and calculate the
        cdf. Apply the inverse normal distribution to the cdf to get the SPI
        SPEI. This process is best described in Lloyd-Hughes and Saunders, 2002
        which is included in the documentation.

        '''

        # Calculate the CDF of observed precipitation on a given time scale
        if not (p_zero is None):
            if params:
                cdf = p_zero + (1 - p_zero) * self.distrb.cdf(data, **params)
            else:
                cdf = np.empty(data.shape)
                cdf.fill(np.nan)
        else:
            cdf = self.distrb.cdf(data, **params)

        # Apply inverse normal distribution
        norm_ppf = scs.norm.ppf(cdf)
        norm_ppf[np.isinf(norm_ppf)] = np.nan

        return norm_ppf

    def calculate(self, df: pd.DataFrame, date_col: str, precip_cols: list, freq: str="M",
                  scale: int=1, freq_col: str=None, fit_type: str='lmom', dist_type: str='gam',
                  **dist_kwargs) -> pd.DataFrame:
        '''
        Calculate the index.

        Check https://docs.scipy.org/doc/scipy/reference/stats.html for
        distribution types

        Parameters
        ----------
        df: pd.Dataframe
            Pandas dataframe with precipitation data as columns. Each column is treated as a seperate
            set of observations and distributions are fit for individual columns. A date column should
            also be given in the dataframe.

        date_col: str
            The column name for the date column. Date specification should follow the strftime format.

        precip_cols: list
            List of columns with precipitation data. Each column is treated as a separate set of
            observations.

        freq: str ["M", "W", "D"]
            The temporal frequency to calculate the index on. The day of year ("D") or week of year
            ("W") or month of year ("M") is derived from the date_col. If the user desires a custome
            frequency such as 3-month, 6-month, they can pass the column name for the custome freqency
            (freq_col)

        freq_col: str (column type: int)
            Name of the column that specifies a custome frequency. This overrides the freq parameter.
            The freq_col should group individual observations (rows) according to the users custome
            frequency. The grouping is specified using integers.

        scale: int (default=1)
            Integer to specify the number of time periods over which the standardized precipitation
            index is to be calculated. If freq="M" then this is the number of months.

        fit_type: str ("lmom" or "mle")
            Specify the type of fit to use for fitting distribution to the precipitation data. Either
            L-moments (lmom) or Maximum Likelihood Estimation (mle). Note use L-moments when comparing
            to NCAR's NCL code and R's packages to calculate SPI and SPEI.

        dist_type: str
            The distribution type to fit using either L-moments or MLE
                'gam' - Gamma
                'exp' - Exponential
                'gev' - Generalised Extreme Value
                'gpa' - Generalised Pareto
                'gum' - Gumbel
                'nor' - Normal
                'pe3' - Pearson III
                'wei' - Weibull

            The distribution type to fit using ONLY MLE
                'glo' - Generalised Logistic
                'gno' - Generalised Normal
                'kap' - Kappa

            The distribution type to fit using ONLY L-moments
                'wak' - Wakeby

        dist_kwargs:
            scale and location parameters. See documentation on scipy.stats.rv_continuous.fit

        Returns
        -------
        df: pd.Dataframe
            Pandas dataframe with the calculated indicies for each precipitation column appended
            to the original dataframe.
        '''

        # Check for duplicate dates
        df = self.check_duplicate_dates(df, date_col)
        if isinstance(precip_cols, str):
            precip_cols = [precip_cols]

        if scale > 1:
            df, precip_cols = self.rolling_window_sum(df, precip_cols, scale)

        self._df_copy = df[[date_col] + precip_cols].copy()
        self._df_copy[date_col] = pd.to_datetime(self._df_copy[date_col])

        if freq_col:
            self.freq_col = freq_col
        else:
            self.freq_col = 'freq'

            if freq == "D":
                self._df_copy[self.freq_col] = self._df_copy[date_col].dt.dayofyear
            elif freq == "W":
                self._df_copy[self.freq_col] = self._df_copy[date_col].dt.week
            elif freq == "M":
                self._df_copy[self.freq_col] = self._df_copy[date_col].dt.month
            else:
                raise AttributeError(f"{freq} is not a recognized frequency. Options are 'M', 'W', or 'D'")

        freq_range = self._df_copy[self.freq_col].unique().tolist()
        # Loop over months
        dfs = []
        for p in precip_cols:
            dfs_p = pd.DataFrame()
            for j in freq_range:
                precip_all = self._df_copy.loc[self._df_copy[self.freq_col]==j]
                precip_single_df = precip_all.dropna().copy()
                precip_single = precip_single_df[p].values
                precip_sorted = np.sort(precip_single)[::-1]

                # Fit distribution for particular series and month
                params, p_zero = self.fit_distribution(
                    precip_sorted, dist_type, fit_type, **dist_kwargs
                )

                # Calculate SPI/SPEI
                spi = self.cdf_to_ppf(precip_single, params, p_zero)
                idx_col = f"{p}_calculated_index"
                precip_single_df[idx_col] = spi
                precip_single_df = precip_single_df[[date_col, idx_col]]
                dfs_p = pd.concat([dfs_p, precip_single_df])
                dfs_p = dfs_p.sort_values(date_col)
            dfs.append(dfs_p)

        df_all = reduce(
            lambda left, right: pd.merge(left, right, on=date_col, how='left'), dfs, self._df_copy
        )
        df_all = df_all.drop(columns=self.freq_col)

        return df_all



import numpy as np
import pandas as pd
from standard_precip import spi


def test_gamma_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="gam"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.678092, 4)

def test_gamma_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="gam", floc=0
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.696543, 4)

def test_exp_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="exp"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.575136, 4)

def test_exp_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="exp"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.177047, 4)

def test_gev_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="gev"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.695562, 4)

def test_gev_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="gev"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.701048, 4)

def test_gpa_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="gpa"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.560434, 4)

def test_gpa_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="gpa"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.288574, 4)

def test_gum_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="gum"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.698605, 4)

def test_gum_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="gum"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.720206, 4)

def test_nor_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="nor"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.753621, 4)

def test_nor_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="nor"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.712801, 4)

def test_pe3_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="pe3"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.652750, 4)

def test_pe3_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="pe3"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.670490, 4)

def test_wei_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="wei"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.624138, 4)

def test_wei_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="wei"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.625167, 4)

def test_glo_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="glo"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.721554, 4)

def test_gno_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="gno"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.655203, 4)

def test_kap_mle():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="mle", dist_type="kap"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[0], 4) == np.round(-0.312152, 4)

def test_wak_lmoments():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/monthly_data.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'TotalPrecipitation', freq="M", fit_type="lmom", dist_type="wak"
    )
    assert np.round(df_spi['TotalPrecipitation_calculated_index'].iloc[3], 4) == np.round(-0.204953, 4)

def test_3month_spi():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/wichita_rain.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'precip', freq="M", fit_type="lmom", scale=3,
        dist_type="gam"
    )
    assert np.round(df_spi['precip_scale_3_calculated_index'].iloc[2], 4) == np.round(0.856479, 4)

def test_daily_nan():
    new_spi = spi.SPI()
    df_rainfall = pd.read_csv('data/daily_data_test.csv')
    df_spi = new_spi.calculate(
        df_rainfall, 'date', 'precip', freq="D", fit_type="lmom", scale=1,
        dist_type="gam"
    )
    assert np.isnan(df_spi['precip_calculated_index'].iloc[0])



# -*- coding: utf-8 -*-

# lmoments3 library
# Copyright (C) 2012, 2014  J. R. M. Hosking, William Asquith,
# Sam Gillespie, Pierre Gérard-Marchant, Florenz A. P. Hollebrandse
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
import numpy as np
import scipy.stats
import scipy.stats._continuous_distns
from scipy import special
import math
from .lmoments_base import lmom_ratios_fn


class LmomDistrMixin(object):
    """
    Mixin class to add L-moment methods to :class:`scipy.stats.rv_continous` distribution functions. Distributions using
    the mixin should override the methods :meth:`._lmom_fit` and :meth:`.lmom_ratios`.
    """

    def _lmom_fit(self, lmom_ratios):
        raise NotImplementedError

    def _lmom_ratios(self, *shapes, locs, scale, nmom):
        """
        When overriding, *shapes can be replaced by the actual distribution's shape parameter(s), if any.
        """
        raise NotImplementedError

    def lmom_fit(self, data=[], lmom_ratios=[]):
        """
        Fit the distribution function to the given data or given L-moments.
        :param data: Data to use in calculating the distribution parameters
        :type data: array_like
        :param lmom_ratios: L-moments (ratios) l1, l2, t3, t4, .. to use in calculating the distribution parameters
        :type lmom_ratios: array_like
        :returns: Distribution parameters in `scipy` order, e.g. scale, loc, shape
        :rtype: :class:`OrderedDict`
        """
        n_min = self.numargs + 2
        if len(data) > 0:
            if len(data) <= n_min:
                raise ValueError("At least {} data points must be provided.".format(n_min))
            lmom_ratios = lmom_ratios_fn(data, nmom=n_min)
        elif not lmom_ratios:
            raise Exception("Either `data` or `lmom_ratios` must be provided.")
        elif len(lmom_ratios) < n_min:
            raise ValueError("At least {} number of L-moments must be provided.".format(n_min))

        return self._lmom_fit(lmom_ratios)

    def lmom(self, *args, nmom=5, **kwds):
        """
        Compute the distribution's L-moments, e.g. l1, l2, l3, l4, ..
        :param args: Distribution parameters in order of shape(s), loc, scale
        :type args: float
        :param nmom: Number of moments to calculate
        :type nmom: int
        :param kwds: Distribution parameters as named arguments. See :attr:`rv_continous.shapes` for names of shape
                     parameters
        :type kwds: float
        :returns: List of L-moments
        :rtype: list
        """
        ratios = self.lmom_ratios(*args, nmom=nmom, **kwds)
        moments = ratios[0:2]
        moments += [ratio * moments[1] for ratio in ratios[2:]]
        return moments

    def lmom_ratios(self, *args, nmom=5, **kwds):
        """
        Compute the distribution's L-moment ratios, e.g. l1, l2, t3, t4, ..
        :param args: Distribution parameters in order of shape(s), loc, scale
        :type args: float
        :param nmom: Number of moments to calculate
        :type nmom: int
        :param kwds: Distribution parameters as named arguments. See :attr:`rv_continous.shapes` for names of shape
                     parameters
        :type kwds: float
        :returns: List of L-moment ratios
        :rtype: list
        """
        if nmom > 20:
            return ValueError("Parameter nmom too large. Max of 20.")

        shapes, loc, scale = self._parse_args(*args, **kwds)

        if scale <= 0:
            return ValueError("Invalid scale parameter.")

        return self._lmom_ratios(*shapes, loc=loc, scale=scale, nmom=nmom)

    def nnlf(self, data, *args, **kwds):
        # Override `nnlf` to provide a more consistent interface with shape and loc and scale parameters

        data = np.asarray(data)
        shapes, loc, scale = self._parse_args(*args, **kwds)
        # This is how scipy's nnlf requires parameters
        theta = list(shapes) + [loc, scale]

        # Now call the super class's nnlf
        return scipy.stats.rv_continuous.nnlf(self, x=data, theta=theta)

    def freeze(self, *args, **kwds):
        # Override `freeze` because we're extending the frozen version of the distribution.
        return LmomFrozenDistr(self, *args, **kwds)


class LmomFrozenDistr(scipy.stats.distributions.rv_frozen):
    """
    Frozen version of the distribution returned by :class:`LmomDistrMixin`. Simply provides additional methods supported
    by the mixin.
    """

    def __init__(self, dist, *args, **kwds):
        super().__init__(dist, *args, **kwds)

    def lmom(self, nmom=5):
        return self.dist.lmom(*self.args, nmom=nmom, **self.kwds)

    def lmom_ratios(self, nmom=5):
        return self.dist.lmom_ratios(*self.args, nmom=nmom, **self.kwds)


"""
The following distributions are **not** available in :mod:`scipy.stats`.
"""

class WakebyGen(LmomDistrMixin, scipy.stats.rv_continuous):
    """
    The Wakeby distribution is defined by the transformation:
    (x-xi)/a = (1/b).[1 - (1-U)^b] - (c/d).[1 - (1-U)^(-d)]
    """

    def _argcheck(self, b, c, d):
        b = np.asarray(b)
        c = np.asarray(c)
        d = np.asarray(d)
        check = np.where(b + d > 0,
                         np.where(c == 0, d == 0, True),
                         (b == c) & (c == d) & (d == 0))
        np.putmask(check, c > 0, d > 0)
        np.putmask(check, c < 0, False)
        return check

    def _ppf(self, q, b, c, d):
        z = -np.log(1. - q)
        u = np.where(b == 0, z, (1. - np.exp(-b * z)) / b)
        v = np.where(d == 0, z, (1. - np.exp(d * z)) / d)
        return u - c * v

    def _cdf(self, x, b, c, d):
        if hasattr(x, '__iter__'):
            if hasattr(b, '__iter__'):
                # Assume x, b, c, d are arrays with matching length
                result = np.array([self._cdfwak(_, parameters)
                                   for (_, parameters) in zip(x, zip(b, c, d))])
            else:
                # Only x is an array, paras are scalars
                result = np.array([self._cdfwak(_, [b, c, d])
                                   for _ in x])
        else:
            result = self._cdfwak(x, (b, c, d))
        return result

    def _cdfwak(self, x, para):
        # Only for a single value of x!

        EPS = 1e-8
        MAXIT = 20
        ZINCMX = 3
        ZMULT = 0.2
        UFL = -170
        XI = 0  # stats.rv_continuous deals with scaling
        A = 1  # stats.rv_continuous deals with scaling
        B, C, D = para

        CDFWAK = 0
        if x <= XI:
            return CDFWAK

        # Test for _special cases
        if B == 0 and C == 0 and D == 0:
            Z = (x - XI) / A
            CDFWAK = 1
            if -Z >= UFL:
                CDFWAK = 1 - math.exp(-Z)
            return CDFWAK

        if C == 0:
            CDFWAK = 1
            if x >= (XI + A / B):
                return (CDFWAK)
            Z = -math.log(1 - (x - XI) * B / A) / B
            if -Z >= UFL:
                CDFWAK = 1 - math.exp(-Z)
            return CDFWAK

        if A == 0:
            Z = math.log(1 + (x - XI) * D / C) / D
            if -Z >= UFL:
                CDFWAK = 1 - math.exp(-Z)
            return CDFWAK

        CDFWAK = 1
        if D < 0 and x >= (XI + A / B - C / D):
            return CDFWAK

        Z = 0.7
        if x < self._ppf(0.1, *para):
            Z = 0
        if x < self._ppf(0.99, *para):
            pass
        else:
            if D < 0:
                Z = math.log((x - XI - A / B) * D / C + 1) / D
            if D == 0:
                Z = (x - XI - A / B) / C
            if D > 0:
                Z = math.log((x - XI) * D / C + 1) / D

        for IT in range(1, MAXIT + 1):
            EB = 0
            BZ = -B * Z
            if BZ >= UFL:
                EB = math.exp(BZ)
            GB = Z

            if abs(B) > EPS:
                GB = (1 - EB) / B
            ED = math.exp(D * Z)
            GD = -Z

            if abs(D) > EPS:
                GD = (1 - ED) / D

            XEST = XI + A * GB - C * GD
            FUNC = x - XEST
            DERIV1 = A * EB + C * ED
            DERIV2 = -A * B * EB + C * D * ED
            TEMP = DERIV1 + 0.5 * FUNC * DERIV2 / DERIV1

            if TEMP <= 0:
                TEMP = DERIV1
            ZINC = FUNC / TEMP

            if ZINC > ZINCMX:
                ZINC = ZINCMX

            ZNEW = Z + ZINC

            if ZNEW <= 0:
                Z = Z * ZMULT
            else:
                Z = ZNEW
                if abs(ZINC) <= EPS:
                    CDFWAK = 1
                    if -Z >= UFL:
                        CDFWAK = 1 - math.exp(-Z)
                    return CDFWAK

    def _pdf(self, x, b, c, d):
        t = (1. - self._cdf(x, b, c, d))
        f = t ** (d + 1) / (t ** (b + d) + c)
        return f

    def _lmom_fit(self, lmom_ratios):
        if lmom_ratios[1] <= 0 or abs(lmom_ratios[2]) >= 1 or abs(lmom_ratios[3]) >= 1 or abs(lmom_ratios[4]) >= 1:
            raise ValueError("Invalid L-Moments")

        ALAM1 = lmom_ratios[0]
        ALAM2 = lmom_ratios[1]
        ALAM3 = lmom_ratios[2] * ALAM2
        ALAM4 = lmom_ratios[3] * ALAM2
        ALAM5 = lmom_ratios[4] * ALAM2

        XN1 = 3 * ALAM2 - 25 * ALAM3 + 32 * ALAM4
        XN2 = -3 * ALAM2 + 5 * ALAM3 + 8 * ALAM4
        XN3 = 3 * ALAM2 + 5 * ALAM3 + 2 * ALAM4
        XC1 = 7 * ALAM2 - 85 * ALAM3 + 203 * ALAM4 - 125 * ALAM5
        XC2 = -7 * ALAM2 + 25 * ALAM3 + 7 * ALAM4 - 25 * ALAM5
        XC3 = 7 * ALAM2 + 5 * ALAM3 - 7 * ALAM4 - 5 * ALAM5

        XA = XN2 * XC3 - XC2 * XN3
        XB = XN1 * XC3 - XC1 * XN3
        XC = XN1 * XC2 - XC1 * XN2
        DISC = XB * XB - 4 * XA * XC
        skip20 = False
        if DISC >= 0:
            DISC = math.sqrt(DISC)
            ROOT1 = 0.5 * (-XB + DISC) / XA
            ROOT2 = 0.5 * (-XB - DISC) / XA
            B = max(ROOT1, ROOT2)
            D = -min(ROOT1, ROOT2)
            if D < 1:
                A = (1 + B) * (2 + B) * (3 + B) / (4 * (B + D)) * ((1 + D) * ALAM2 - (3 - D) * ALAM3)
                C = -(1 - D) * (2 - D) * (3 - D) / (4 * (B + D)) * ((1 - B) * ALAM2 - (3 + B) * ALAM3)
                XI = ALAM1 - A / (1 + B) - C / (1 - D)
                skip20 = bool(C >= 0 and A + C >= 0)

        if not skip20:
            D = -(1 - 3 * lmom_ratios[2]) / (1 + lmom_ratios[2])
            C = (1 - D) * (2 - D) * lmom_ratios[1]
            B = 0
            A = 0
            XI = lmom_ratios[0] - C / (1 - D)
            if D <= 0:
                A = C
                B = -D
                C = 0
                D = 0

        para = OrderedDict([('beta', B),
                            ('gamma', C),
                            ('delta', D),
                            ('loc', XI),
                            ('scale', A)])
        return para

    def _lmom_ratios(self, beta, gamma, delta, loc, scale, nmom):
        if (delta >= 1) \
                or (beta + delta <= 0 and (beta != 0 or gamma != 0 or delta != 0)) \
                or (scale == 0 and beta != 0) \
                or (gamma == 0 and delta != 0) \
                or (gamma < 0) \
                or (scale + gamma < 0) \
                or (scale == 0 and gamma == 0):
            raise ValueError("Invalid parameters")

        Y = scale / (1 + beta)
        Z = gamma / (1 - delta)
        xmom = [loc + Y + Z]
        if nmom == 1:
            return xmom

        Y /= (2 + beta)
        Z /= (2 - delta)
        ALAM2 = Y + Z
        xmom.append(ALAM2)
        if nmom == 2:
            return xmom

        for i in range(2, nmom):
            Y *= (i - 1 - beta) / (i + 1 + beta)
            Z *= (i - 1 + delta) / (i + 1 - delta)
            xmom.append((Y + Z) / ALAM2)
        return xmom


wak = WakebyGen(name='wakeby', shapes='beta, gamma, delta')

"""
The following distributions are available in `scipy.stats` and are redefined here with an `LmomDistrMixin` to extend the
scipy distribution with L-moment methods.
"""

class GenParetoGen(LmomDistrMixin, scipy.stats._continuous_distns.genpareto_gen):
    def _lmom_fit(self, lmom_ratios):
        T3 = lmom_ratios[2]
        if lmom_ratios[1] <= 0 or abs(T3) >= 1:
            raise ValueError("L-Moments invalid")

        G = (1 - 3 * T3) / (1 + T3)

        # CHANGE: shape parameter `c` has been negated from original `lmoments` package to be compatible with scipy's
        # GPA distribution function.
        PARA3 = -G
        PARA2 = (1 + G) * (2 + G) * lmom_ratios[1]
        PARA1 = lmom_ratios[0] - PARA2 / (1 + G)

        para = OrderedDict([('c', PARA3),
                            ('loc', PARA1),
                            ('scale', PARA2)])
        return para

    def _lmom_ratios(self, c, loc, scale, nmom):
        # See above, shape parameter negated
        G = -c
        if c > 1:
            raise ValueError("Invalid Parameters")

        Y = 1 / (1 + G)
        xmom = [loc + scale * Y]
        if nmom == 1:
            return xmom

        Y /= (2 + G)
        xmom.append(scale * Y)
        if nmom == 2:
            return xmom

        Y = 1
        for i in range(3, nmom + 1):
            Y *= (i - 2 - G) / (i + G)
            xmom.append(Y)
        return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('c', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


gpa = GenParetoGen(a=0.0, name='genpareto', shapes='c')


class ExponGen(LmomDistrMixin, scipy.stats._continuous_distns.expon_gen):
    def _lmom_fit(self, lmom_ratios):
        if lmom_ratios[1] <= 0:
            raise ValueError("L-Moments invalid")

        para = OrderedDict([('loc', lmom_ratios[0] - 2 * lmom_ratios[1]),
                            ('scale', 2 * lmom_ratios[1])])
        return para

    def _lmom_ratios(self, loc, scale, nmom):
        xmom = [loc + scale]
        if nmom == 1:
            return xmom

        xmom.append(0.5 * scale)
        if nmom == 2:
            return xmom

        for i in range(3, nmom + 1):
            xmom.append(2 / (i * (i - 1)))

        return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('loc', params[0]),
            ('scale', params[1])
        ])
        return param_dict


exp = ExponGen(a=0.0, name='expon')


class GammaGen(LmomDistrMixin, scipy.stats._continuous_distns.gamma_gen):
    def _lmom_fit(self, lmom_ratios):
        A1 = -0.3080
        A2 = -0.05812
        A3 = 0.01765
        B1 = 0.7213
        B2 = -0.5947
        B3 = -2.1817
        B4 = 1.2113

        if lmom_ratios[0] <= lmom_ratios[1] or lmom_ratios[1] <= 0:
            raise ValueError("L-Moments invalid")

        CV = lmom_ratios[1] / lmom_ratios[0]
        if CV >= 0.5:
            T = 1 - CV
            ALPHA = T * (B1 + T * B2) / (1 + T * (B3 + T * B4))
        else:
            T = math.pi * CV ** 2
            ALPHA = (1 + A1 * T) / (T * (1 + T * (A2 + T * A3)))

        para = OrderedDict([('a', ALPHA),
                            ('loc', 0),
                            ('scale', lmom_ratios[0] / ALPHA)])
        return para

    def _lmom_ratios(self, a, loc, scale, nmom):
        A0 = 0.32573501
        [A1, A2, A3] = [0.16869150, 0.078327243, -0.0029120539]
        [B1, B2] = [0.46697102, 0.24255406]
        C0 = 0.12260172
        [C1, C2, C3] = [0.053730130, 0.043384378, 0.011101277]
        [D1, D2] = [0.18324466, 0.20166036]
        [E1, E2, E3] = [2.3807576, 1.5931792, 0.11618371]
        [F1, F2, F3] = [5.1533299, 7.1425260, 1.9745056]

        if a <= 0:
            raise ValueError("Invalid Parameters")
        if nmom > 4:
            raise ValueError("Parameter nmom too large")
        if loc != 0:
            raise ValueError("Location parameter not supported for Gamma distribution.")

        xmom = []
        xmom.append(a * scale)
        if nmom == 1:
            return (xmom)

        xmom.append(scale * 1 / math.sqrt(math.pi) * math.exp(special.gammaln(a + 0.5) - special.gammaln(a)))
        if nmom == 2:
            return (xmom)

        if a < 1:
            Z = a
            xmom.append((((E3 * Z + E2) * Z + E1) * Z + 1) / (((F3 * Z + F2) * Z + F1) * Z + 1))
            if nmom == 3:
                return (xmom)
            xmom.append((((C3 * Z + C2) * Z + C1) * Z + C0) / ((D2 * Z + D1) * Z + 1))
            if nmom == 4:
                return (xmom)
        else:
            Z = 1 / a
            xmom.append(math.sqrt(Z) * (((A3 * Z + A2) * Z + A1) * Z + A0) / ((B2 * Z + B1) * Z + 1))
            if nmom == 3:
                return (xmom)

            xmom.append((((C3 * Z + C2) * Z + C1) * Z + C0) / ((D2 * Z + D1) * Z + 1))
            if nmom == 4:
                return (xmom)

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('a', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


gam = GammaGen(a=0.0, name='gamma', shapes='a')


class GenextremeGen(LmomDistrMixin, scipy.stats._continuous_distns.genextreme_gen):
    def _lmom_fit(self, lmom_ratios):
        SMALL = 1e-5
        eps = 1e-6
        maxit = 20
        EU = 0.57721566
        DL2 = math.log(2)
        DL3 = math.log(3)
        A0 = 0.28377530
        A1 = -1.21096399
        A2 = -2.50728214
        A3 = -1.13455566
        A4 = -0.07138022
        B1 = 2.06189696
        B2 = 1.31912239
        B3 = 0.25077104
        C1 = 1.59921491
        C2 = -0.48832213
        C3 = 0.01573152
        D1 = -0.64363929
        D2 = 0.08985247

        T3 = lmom_ratios[2]
        if lmom_ratios[1] <= 0 or abs(T3) >= 1:
            raise ValueError("L-Moments Invalid")

        if T3 <= 0:
            G = (A0 + T3 * (A1 + T3 * (A2 + T3 * (A3 + T3 * A4)))) / (1 + T3 * (B1 + T3 * (B2 + T3 * B3)))

            if T3 >= -0.8:
                para3 = G
                GAM = math.exp(special.gammaln(1 + G))
                para2 = lmom_ratios[1] * G / (GAM * (1 - 2 ** -G))
                para1 = lmom_ratios[0] - para2 * (1 - GAM) / G
                para = OrderedDict([('c', para3),
                                    ('loc', para1),
                                    ('scale', para2)])
                return para
            elif T3 <= -0.97:
                G = 1 - math.log(1 + T3) / DL2

            T0 = (T3 + 3) * 0.5
            for IT in range(1, maxit):
                X2 = 2 ** -G
                X3 = 3 ** -G
                XX2 = 1 - X2
                XX3 = 1 - X3
                T = XX3 / XX2
                DERIV = (XX2 * X3 * DL3 - XX3 * X2 * DL2) / (XX2 ** 2)
                GOLD = G
                G -= (T - T0) / DERIV

                if abs(G - GOLD) <= eps * G:
                    para3 = G
                    GAM = math.exp(special.gammaln(1 + G))
                    para2 = lmom_ratios[1] * G / (GAM * (1 - 2 ** -G))
                    para1 = lmom_ratios[0] - para2 * (1 - GAM) / G
                    para = OrderedDict([('c', para3),
                                        ('loc', para1),
                                        ('scale', para2)])
                    return para
            raise Exception("Iteration has not converged")
        else:
            Z = 1 - T3
            G = (-1 + Z * (C1 + Z * (C2 + Z * C3))) / (1 + Z * (D1 + Z * D2))
            if abs(G) < SMALL:
                para2 = lmom_ratios[1] / DL2
                para1 = lmom_ratios[0] - EU * para2
                para = OrderedDict([('c', 0),
                                    ('loc', para1),
                                    ('scale', para2)])
            else:
                para3 = G
                GAM = math.exp(special.gammaln(1 + G))
                para2 = lmom_ratios[1] * G / (GAM * (1 - 2 ** -G))
                para1 = lmom_ratios[0] - para2 * (1 - GAM) / G
                para = OrderedDict([('c', para3),
                                    ('loc', para1),
                                    ('scale', para2)])
            return para

    def _lmom_ratios(self, c, loc, scale, nmom):
        ZMOM = [0.577215664901532861, 0.693147180559945309,
                0.169925001442312363, 0.150374992788438185,
                0.558683500577583138e-1, 0.581100239999710876e-1,
                0.276242584297309125e-1, 0.305563766579053126e-1,
                0.164650282258328802e-1, 0.187846624298170912e-1,
                0.109328215063027148e-1, 0.126973126676329530e-1,
                0.778982818057231804e-2, 0.914836179621999726e-2,
                0.583332389328363588e-2, 0.690104287590348154e-2,
                0.453267970180679549e-2, 0.538916811326595459e-2,
                0.362407767772390e-2, 0.432387608605538096e-2]
        SMALL = 1e-6
        if c <= -1:
            raise ValueError("Invalid Parameters")

        if abs(c) > SMALL:
            GAM = math.exp(special.gammaln(1 + c))
            xmom = [loc + scale * (1 - GAM) / c]
            if nmom == 1:
                return xmom

            XX2 = 1 - 2 ** (-c)
            xmom.append(scale * XX2 * GAM / c)
            if nmom == 2:
                return xmom

            Z0 = 1
            for j in range(2, nmom):
                DJ = j + 1
                BETA = (1 - DJ ** -c) / XX2
                Z0 *= (4 * DJ - 6) / DJ
                Z = Z0 * 3 * (DJ - 1) / (DJ + 1)
                SUM = Z0 * BETA - Z
                if j == 2:
                    xmom.append(SUM)
                else:
                    for i in range(1, j - 1):
                        DI = i + 1
                        Z *= (DI + DI + 1) * (DJ - DI) / ((DI + DI - 1) * (DJ + DI))
                        SUM -= Z * xmom[i + 1]
                    xmom.append(SUM)
            return xmom

        else:
            xmom = [loc]
            if nmom == 1:
                return xmom

            xmom.append(scale * ZMOM[1])
            if nmom == 2:
                return xmom

            for i in range(2, nmom):
                xmom.append(ZMOM[i - 1])

            return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('c', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


gev = GenextremeGen(name='genextreme', shapes='c')


class GumbelGen(LmomDistrMixin, scipy.stats._continuous_distns.gumbel_r_gen):
    def _lmom_fit(self, lmom_ratios):
        EU = 0.577215664901532861
        if lmom_ratios[1] <= 0:
            raise ValueError("L-Moments Invalid")

        para2 = lmom_ratios[1] / math.log(2)
        para1 = lmom_ratios[0] - EU * para2
        para = OrderedDict([('loc', para1),
                            ('scale', para2)])
        return para

    def _lmom_ratios(self, loc, scale, nmom):
        ZMOM = [0.577215664901532861, 0.693147180559945309, 0.169925001442312363, 0.150374992788438185,
                0.0558683500577583138, 0.0581100239999710876, 0.0276242584297309125, 0.0305563766579053126,
                0.0164650282258328802, 0.0187846624298170912, 0.0109328215063027148, 0.0126973126676329530,
                0.00778982818057231804, 0.00914836179621999726, 0.00583332389328363588, 0.00690104287590348154,
                0.00453267970180679549, 0.00538916811326595459, 0.00362407767772368790, 0.00432387608605538096]
        xmom = [loc + scale * ZMOM[0]]
        if nmom == 1:
            return xmom

        xmom.append(scale * ZMOM[1])
        if nmom == 2:
            return xmom

        for i in range(2, nmom):
            xmom.append(ZMOM[i])
        return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('loc', params[0]),
            ('scale', params[1])
        ])
        return param_dict


gum = GumbelGen(name='gumbel_r')


class NormGen(LmomDistrMixin, scipy.stats._continuous_distns.norm_gen):
    def _lmom_fit(self, lmom_ratios):
        if lmom_ratios[1] <= 0:
            raise ValueError("L-Moments invalid")

        para = OrderedDict([('loc', lmom_ratios[0]),
                            ('scale', lmom_ratios[1] * math.sqrt(math.pi))])
        return para

    def _lmom_ratios(self, *shapes, loc, scale, nmom):
        ZMOM = [0, 0.564189583547756287, 0, 0.122601719540890947, 0, 0.0436611538950024944, 0, 0.0218431360332508776, 0,
                0.0129635015801507746, 0, 0.00852962124191705402, 0, 0.00601389015179323333, 0, 0.00445558258647650150,
                0, 0.00342643243578076985, 0, 0.00271267963048139365]
        xmom = [loc]
        if nmom == 1:
            return xmom

        xmom.append(scale * ZMOM[1])
        if nmom == 2:
            return xmom

        for M in range(2, nmom):
            xmom.append(ZMOM[M])

        return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('loc', params[0]),
            ('scale', params[1])
        ])
        return param_dict


nor = NormGen(name='norm')


class Pearson3Gen(LmomDistrMixin, scipy.stats._continuous_distns.pearson3_gen):
    def _lmom_fit(self, lmom_ratios):
        Small = 1e-6
        # Constants used in Minimax Approx:
        C1 = 0.2906
        C2 = 0.1882
        C3 = 0.0442
        D1 = 0.36067
        D2 = -0.59567
        D3 = 0.25361
        D4 = -2.78861
        D5 = 2.56096
        D6 = -0.77045

        T3 = abs(lmom_ratios[2])
        if lmom_ratios[1] <= 0 or T3 >= 1:
            raise ValueError("L-Moments invalid")

        if T3 <= Small:
            loc = lmom_ratios[0]
            scale = lmom_ratios[1] * math.sqrt(math.pi)
            skew = 0
        else:
            if T3 >= (1.0 / 3):
                T = 1 - T3
                Alpha = T * (D1 + T * (D2 + T * D3)) / (1 + T * (D4 + T * (D5 + T * D6)))
            else:
                T = 3 * math.pi * T3 * T3
                Alpha = (1 + C1 * T) / (T * (1 + T * (C2 + T * C3)))

            RTALPH = math.sqrt(Alpha)
            BETA = math.sqrt(math.pi) * lmom_ratios[1] * math.exp(special.gammaln(Alpha) - special.gammaln(Alpha + 0.5))
            loc = lmom_ratios[0]
            scale = BETA * RTALPH
            skew = 2 / RTALPH
            if lmom_ratios[2] < 0:
                skew *= -1

        return OrderedDict([('skew', skew),
                            ('loc', loc),
                            ('scale', scale)])

    def _lmom_ratios(self, skew, loc, scale, nmom):
        SMALL = 1e-6
        CONST = 1 / math.sqrt(math.pi)
        A0 = 0.32573501
        [A1, A2, A3] = [0.16869150, 0.078327243, -0.0029120539]
        [B1, B2] = [0.46697102, 0.24255406]
        C0 = 0.12260172
        [C1, C2, C3] = 0.053730130, 0.043384378, 0.011101277
        [D1, D2] = [0.18324466, 0.20166036]
        [E1, E2, E3] = [2.3807576, 1.5931792, 0.11618371]
        [F1, F2, F3] = [5.1533299, 7.1425260, 1.9745056]
        [G1, G2, G3] = [2.1235833, 4.1670213, 3.1925299]
        [H1, H2, H3] = [9.0551443, 26.649995, 26.193668]

        # Calculate only up to 4 L-moments for Pearson 3
        if nmom > 4:
            raise ValueError("Parameter nmom too large")

        xmom = [loc]
        if nmom == 1:
            return xmom

        if abs(skew) < SMALL:
            xmom = [loc]
            if nmom == 1:
                return xmom

            xmom.append(CONST * scale)
            if nmom == 2:
                return xmom

            xmom.append(0)
            if nmom == 3:
                return xmom

            xmom.append(C0)
            return xmom
        else:
            Alpha = 4 / (skew * skew)
            Beta = abs(0.5 * scale * skew)
            ALAM2 = CONST * math.exp(special.gammaln(Alpha + 0.5) - special.gammaln(Alpha))
            xmom.append(ALAM2 * Beta)
            if nmom == 2:
                return xmom

            if Alpha < 1:
                Z = Alpha
                xmom.append((((E3 * Z + E2) * Z + E1) * Z + 1) / (((F3 * Z + F2) * Z + F1) * Z + 1))
                if skew < 0:
                    xmom[2] *= -1
                if nmom == 3:
                    return xmom

                xmom.append((((G3 * Z + G2) * Z + G1) * Z + 1) / (((H3 * Z + H2) * Z + H1) * Z + 1))
                return xmom
            else:
                Z = 1.0 / Alpha
                xmom.append(math.sqrt(Z) * (((A3 * Z + A2) * Z + A1) * Z + A0) / ((B2 * Z + B1) * Z + 1))
                if skew < 0:
                    xmom[2] *= -1
                if nmom == 3:
                    return xmom

                xmom.append((((C3 * Z + C2) * Z + C1) * Z + C0) / ((D2 * Z + D1) * Z + 1))
                return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('skew', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


pe3 = Pearson3Gen(name="pearson3", shapes='skew')


class WeibullMinGen(LmomDistrMixin, scipy.stats._continuous_distns.weibull_min_gen):
    def _lmom_fit(self, lmom_ratios):
        if lmom_ratios[1] <= 0 or lmom_ratios[2] >= 1 or lmom_ratios[2] <= -gum.lmom_ratios(nmom=3)[2]:
            raise ValueError("L-Moments invalid")

        pg = gev.lmom_fit(lmom_ratios=[-lmom_ratios[0], lmom_ratios[1], -lmom_ratios[2]])
        delta = 1 / pg['c']
        beta = pg['scale'] / pg['c']
        para = OrderedDict([('c', delta),
                            ('loc', -pg['loc'] - beta),
                            ('scale', beta)])
        return para

    def _lmom_ratios(self, c, loc, scale, nmom):
        if c <= 0:
            raise ValueError("Invalid parameters")

        xmom = gev.lmom_ratios(loc=0, scale=scale / c, c=1 / c, nmom=nmom)
        xmom[0] = loc + scale - xmom[0]
        xmom[2] *= -1
        return xmom

    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('c', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


wei = WeibullMinGen(a=0.0, name='weibull_min', shapes='c')


class GenLogistic(LmomDistrMixin, scipy.stats._continuous_distns.genlogistic_gen):
    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('c', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


glo = GenLogistic(name='genlogistic')


class GenNormal(LmomDistrMixin, scipy.stats._continuous_distns.gennorm_gen):
    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('beta', params[0]),
            ('loc', params[1]),
            ('scale', params[2])
        ])
        return param_dict


gno = GenNormal(name='gennorm')


class Kappa4(LmomDistrMixin, scipy.stats._continuous_distns.kappa4_gen):
    def fit(self, data, *args, **kwargs):
        params = super().fit(data, *args, **kwargs)

        param_dict = OrderedDict([
            ('h', params[0]),
            ('k', params[1]),
            ('loc', params[2]),
            ('scale', params[3])
        ])
        return param_dict


kap = Kappa4(name='kappa4')

